import sys
import unittest

from mock import call, patch

from gapipy.client import Client
from gapipy.request import APIRequestor

from .fixtures import FIRST_PAGE_LIST_DATA, SECOND_PAGE_LIST_DATA


class Resources(object):
    def __init__(self, **kwArgs):
        self.__dict__.update(kwArgs)


class APIRequestorTestCase(unittest.TestCase):

    def setUp(self):
        self.client = Client()
        self.resources = Resources(_resource_name='resources', _uri=None)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_resource_by_id(self, mock_request):
        requestor = APIRequestor(self.client, self.resources)
        requestor.get(1234)
        mock_request.assert_called_once_with('/resources/1234', 'GET', additional_headers=None)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_with_null_resource_id_and_uri_raises_error(self, mock_request):
        requestor = APIRequestor(self.client, self.resources)
        error_msg = 'Need to provide at least one of `resource_id` or `uri` as argument'
        if sys.version_info.major < 3:
            with self.assertRaisesRegexp(ValueError, error_msg):
                requestor.get()
        else:
            with self.assertRaisesRegex(ValueError, error_msg):
                requestor.get()

    @patch('gapipy.request.APIRequestor._request')
    def test_get_with_falsy_resource_id_does_not_raise_error(self, mock_request):
        requestor = APIRequestor(self.client, self.resources)
        requestor.get(0)
        mock_request.assert_called_once_with('/resources/0', 'GET', additional_headers=None)

    @patch('gapipy.request.APIRequestor._request')
    def test_list_resource(self, mock_request):
        requestor = APIRequestor(self.client, self.resources)
        requestor.list_raw()
        mock_request.assert_called_once_with('/resources', 'GET', params=None)

    @patch('gapipy.request.APIRequestor._request')
    def test_list_resource_with_parent(self, mock_request):
        parent = ('parent', '1234', None)
        requestor = APIRequestor(self.client, self.resources, parent=parent)
        requestor.list_raw()
        mock_request.assert_called_once_with(
            '/parent/1234/resources', 'GET', params=None)

    @patch('gapipy.request.APIRequestor.list_raw')
    def test_list_generator(self, mock_list):
        mock_list.side_effect = [FIRST_PAGE_LIST_DATA, SECOND_PAGE_LIST_DATA]
        expected_calls = [
            call(None),
            call('http://localhost:5000/resources/?page=2'),
        ]

        requestor = APIRequestor(self.client, self.resources)
        resources = list(requestor.list())

        self.assertEqual(mock_list.mock_calls, expected_calls)
        self.assertEqual(len(resources), 6)

    @patch('gapipy.request.APIRequestor._request')
    def test_uuid_not_set(self, mock_request):
        self.client.uuid = False
        requestor = APIRequestor(self.client, self.resources)
        requestor.params = {'test': '1234'}
        requestor.list_raw()
        mock_request.assert_called_once_with('/resources', 'GET', params={'test': '1234'})

    @patch('gapipy.request.APIRequestor._make_call')
    def test_uuid_set(self, mock_make_call):
        self.client.uuid = True
        requestor = APIRequestor(self.client, self.resources)
        requestor.list_raw()
        params_arg = mock_make_call.call_args[0][-1]
        self.assertTrue('uuid' in params_arg)

    @patch('gapipy.request.APIRequestor._make_call')
    def test_uuid_with_other_params(self, mock_make_call):
        self.client.uuid = True
        requestor = APIRequestor(self.client, self.resources)
        requestor.params = {'test': '1234'}
        requestor.list_raw()
        params_arg = mock_make_call.call_args[0][-1]
        self.assertEqual(params_arg['test'], '1234')
        self.assertTrue('uuid' in params_arg)



# Python 2 and 3
from __future__ import unicode_literals

from ...models.base import BaseModel
from ..base import Resource

class ServiceLevel(Resource):
    _resource_name = 'service_levels'

    _as_is_fields = [
        'id', 'href', 'code',
        'label',
        'description',
    ]



# TODO: Replace fixtures by factories (or anything more manageable)
# TODO: Replace hardcoded urls with config.API_ROOT?

PPP_TOUR_DATA = {
    'departures': {'href': 'http://localhost:5000/tours/21346/departures'},
    'departures_end_date': '2013-12-31',
    'departures_start_date': '2013-01-01',
    'tour_dossier': {
        'href': 'http://localhost:5000/tour_dossiers/21346',
        'id': '21346'
    },
    'product_line': 'PPP',
    'href': 'http://localhost:5000/tours/21346',
    'id': '21346'
}


TOUR_DOSSIER_LIST_DATA = {
    'count': 3,
    'max_per_page': 3,
    'current_page': 1,
    'results': [
        {
            'href': 'http://localhost:5000/tour_dossiers/1234',
            'id': '1234',
            'tour_dossier_code': 'ABCD'
        },
        {
            'href': 'http://localhost:5000/tour_dossiers/5678',
            'id': '5678',
            'tour_dossier_code': 'EFGH'
        },
        {
            'href': 'http://localhost:5000/tour_dossiers/9012',
            'id': '9012',
            'tour_dossier_code': 'IJKL'
        }
    ],
    'links': []
}

FIRST_PAGE_LIST_DATA = {
    'count': 6,
    'max_per_page': 3,
    'current_page': 1,
    'results': [
        {
            'href': 'http://localhost:5000/resources/1234',
            'id': '1234',
        },
        {
            'href': 'http://localhost:5000/resources/5678',
            'id': '5678',
        },
        {
            'href': 'http://localhost:5000/resources/9012',
            'id': '9012',
        }
    ],
    'links': [
        {
            'href': 'http://localhost:5000/resources/?page=2',
            'rel': 'next',
        }
    ]
}

SECOND_PAGE_LIST_DATA = {
    'count': 6,
    'max_per_page': 3,
    'current_page': 2,
    'results': [
        {
            'href': 'http://localhost:5000/resources/3456',
            'id': '3456',
        },
        {
            'href': 'http://localhost:5000/resources/7890',
            'id': '7890',
        },
        {
            'href': 'http://localhost:5000/resources/1111',
            'id': '1111',
        }
    ],
    'links': [
        {
            'href': 'http://localhost:5000/resources/',
            'rel': 'prev',
        }
    ]
}
PPP_DOSSIER_DATA = {
    "departures_start_date": "2013-01-01",
    "departures_end_date": "2013-12-31",
    "description": "Spot macaws in the jungle and caimans on the riverbanks, sail the waters of Lake Titicaca, delight in the smells of markets and explore ancient ruins, including a trek along the Inca Trail. As we operate our own treks, our quality equipment and the expertise of our porters and CEOs will ensure that your first glimpse of Machu Picchu will leave you in awe. Whether you're scanning the canopy for wildlife from the comfort of our intimate and exclusive G Lodge Amazon or following in the footsteps of the Incas, this fast-paced adventure provides a sweeping view of this diverse country.",
    "tour": {
        "href": "http://localhost:5000/tours/21346",
        "id": "21346"
    },
    "href": "http://localhost:5000/tour_dossiers/21346",
    "product_line": "PPP",
    "images": [
        {
            "image_href": "https://media.gadventures.com/media-server/dynamic/admin/maps/2013/PPP.jpg",
            "type": "MAP"
        },
        {
            "image_href": "https://media.gadventures.com/media-server/dynamic/admin/trip-banners/south-america/peru/PPP/PPP_banner_peru_landscape_ruins_machu_picchu_clouds.jpg",
            "type": "BANNER"
        }
    ],
    "details": [
        {
            "body": "Meeting the locals at a Lake Titicaca homestay, climbing the Inca Trail to Machu Picchu, spotting wildlife at our exclusive G Lodge Amazon, devouring seafood in Lima.",
            "detail_type": {
                "id": "2",
                "label": "Highlights"
            }
        },
        {
            "body": "16",
            "detail_type": {
                "id": "113",
                "label": "Max Pax"
            }
        },
        {
            "body": "Chief Experience Officer (CEO) throughout, specialist Inca Trail CEO on hike.",
            "detail_type": {
                "id": "18",
                "label": "Group Leader"
            }
        }
    ],
    "geography": {
        "start_city": {
            "name": "Lima"
        },
        "start_country": {
            "href": "http://localhost:5000/countries/PE",
            "id": "PE",
            "name": "Peru"
        },
        "region": {
            "name": "South America"
        },
        "visited_countries": [
            {
                "href": "http://localhost:5000/countries/PE",
                "id": "PE",
                "name": "Peru"
            }
        ],
        "finish_country": {
            "href": "http://localhost:5000/countries/PE",
            "id": "PE",
            "name": "Peru"
        },
        "primary_country": {
            "href": "http://localhost:5000/countries/PE",
            "id": "PE",
            "name": "Peru"
        },
        "finish_city": {
            "name": "Lima"
        }
    },
    "id": "21346",
    "categories": [
        {
            "category_type": {
                "id": "1",
                "label": "Activity"
            },
            "name": "Wildlife/Nature"
        },
        {
            "category_type": {
                "id": "1",
                "label": "Activity"
            },
            "name": "Trekking/Hiking"
        },
        {
            "category_type": {
                "id": "1",
                "label": "Activity"
            },
            "name": "Culture/History"
        },
        {
            "category_type": {
                "id": "16",
                "label": "Trip Style"
            },
            "name": "Classic"
        },
        {
            "category_type": {
                "id": "28",
                "label": "Service Level"
            },
            "name": "Standard"
        },
        {
            "category_type": {
                "id": "33",
                "label": "Physical Grading"
            },
            "name": "4"
        },
        {
            "category_type": {
                "id": "39",
                "label": "Merchandising"
            },
            "name": "Top Seller"
        },
        {
            "category_type": {
                "id": "58",
                "label": "Trip Type"
            },
            "name": "Small Group"
        }
    ],
    "name": "Peru Panorama"
}


PRODUCTS_DATA = [
    {
        "href": "https://localhost:5000/departures/1234",
        "id": "1234",
        "type": "departures",
        "sub_type": "Tour",
    },
    {
        "href": "https://localhost:5000/departures/5678",
        "id": "5678",
        "type": "departures",
        "sub_type": "Tour",
    },
    {
        "href": "https://localhost:5000/departures/9012",
        "id": "9012",
        "type": "departures",
        "sub_type": "Tour",
    }
]

DUMMY_DEPARTURE = {
    "sku": "GAPPPP130406-O1",
    "product_line": "PPP",
    "tour_dossier": {
        "href": "https://rest.gadventures.com/tour_dossiers/9882",
        "id": "9882",
        "name": "Peru Panorama"
    },
    "date_last_modified": "2013-06-10T19:12:04Z",
    "finish_date": "2013-04-20",
    "lowest_pp2a_prices": [
        {
            "currency": "USD",
            "amount": "3099.00"
        },
        {
            "currency": "AUD",
            "amount": "2999.00"
        },
        {
            "currency": "CHF",
            "amount": "2949.00"
        },
        {
            "currency": "GBP",
            "amount": "1749.00"
        },
        {
            "currency": "NZD",
            "amount": "3999.00"
        },
        {
            "currency": "CAD",
            "amount": "3299.00"
        },
        {
            "currency": "EUR",
            "amount": "2149.00"
        }
    ],
    "availability": {
        "status": "NOT_BOOKABLE",
        "total": 8
    },
    "tour": {
        "href": "https://rest.gadventures.com/tours/9882",
        "id": "9882"
    },
    "nearest_finish_airport": {
        "code": "LIM"
    },
    "id": "419504",
    "latest_arrival_time": "2013-04-06T23:59:59",
    "href": "https://rest.gadventures.com/departures/419504",
    "flags": [],
    "earliest_departure_time": "2013-04-20T00:00:00",
    "addons": [
        {
            "product": {
                "id": "54",
                "href": "https://rest.gadventures.com/accommodations/54",
                "name": "San Agustin Colonial",
                "type": "accommodations",
                "sub_type": "Hotel"
            },
            "start_date": "2013-04-02",
            "finish_date": "2013-04-06",
            "min_days": 1,
            "max_days": 5,
            "request_space_date": "2013-04-01",
            "halt_booking_date": "2013-04-01",
        },
        {
            "product": {
                "id": "54",
                "href": "https://rest.gadventures.com/accommodations/54",
                "name": "San Agustin Colonial",
                "type": "accommodations",
                "sub_type": "Hotel"
            },
            "start_date": "2013-04-20",
            "finish_date": "2013-04-24",
            "min_days": 1,
            "max_days": 5,
            "request_space_date": "2013-04-01",
            "halt_booking_date": "2013-04-01",
        },
        {
            "product": {
                "id": "298",
                "href": "https://rest.gadventures.com/transports/298",
                "name": "Airport to Hotel Transfer",
                "type": "transports",
                "sub_type": "Transfer"
            },
            "start_date": "2013-04-02",
            "finish_date": "2013-04-06",
            "min_days": 1,
            "max_days": 1,
            "request_space_date": "2013-04-01",
            "halt_booking_date": "2013-04-01",
        },
        {
            "product": {
                "id": "4486",
                "href": "https://rest.gadventures.com/activities/4486",
                "name": "Peru Culinary Theme Pack",
                "type": "activities",
                "sub_type": "Product"
            },
            "start_date": "2013-04-16",
            "finish_date": "2013-04-17",
            "min_days": 1,
            "max_days": 1,
            "request_space_date": "2013-04-01",
            "halt_booking_date": "2013-04-01",
        }
    ],
    "date_created": "2012-01-26T20:29:07Z",
    "nearest_start_airport": {
        "code": "LIM"
    },
    "start_date": "2013-04-06",
    "rooms": [
        {
            "code": "STANDARD",
            "name": "Standard",
            "flags": [],
            "price_bands": [
                {
                    "max_age": 100,
                    "code": "ADULT",
                    "name": "Adult",
                    "max_travellers": 16,
                    "min_age": 12,
                    "prices": [
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "AUD",
                            "amount": "2999.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "CAD",
                            "amount": "3299.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "CHF",
                            "amount": "2949.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "EUR",
                            "amount": "2149.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "GBP",
                            "amount": "1749.00",
                            "deposit": "100.00"
                        },
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "NZD",
                            "amount": "3999.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [
                                {
                                    "amount": "2199.20",
                                    "href": "https://rest.gadventures.com/promotions/48500",
                                    "id": "48500"
                                },
                            ],
                            "currency": "USD",
                            "amount": "3099.00",
                            "deposit": "250.00"
                        }
                    ],
                    "min_travellers": 1
                }
            ],
            "addons": [
                {
                    "product": {
                        "id": "T419504",
                        "href": "https://rest.gadventures.com/single_supplements/T419504",
                        "name": "My Own Room",
                        "type": "single_supplements",
                        "sub_type": "Single Supplement"
                    },
                    "start_date": "2013-04-06",
                    "finish_date": "2013-04-20",
                    "min_days": 14,
                    "max_days": 14,
                    "request_space_date": "2013-04-01",
                    "halt_booking_date": "2013-04-01",
                }
            ],
            "availability": {
                "status": "NOT_BOOKABLE",
                "total": 8,
                "male": None,
                "female": None
            }
        }
    ]
}


DUMMY_PROMOTION = {
    "currencies": ["AUD", "CAD", "CHF", "EUR", "GBP", "NZD", "USD"],
    "name": "10% off North America",
    "commission_rate": 0,
    "sale_finish_date": "2012-08-27",
    "discount_percent": "10.00",
    "id": "18336",
    "terms_and_conditions": "From big cities to untamed wildlife and scenic landscapes, there's a North American adventure for everyone.\r\n<ul><li>Travellers receive 10% off all North America tours (valid for all North America trip codes).</li>\r\n<li>Book by August 27, 2012 for departures through December 16, 2012.</li></ul>\r\nFor general Terms and Conditions please visit our <a href=http://www.gadventures.com/terms-and-conditions/>Terms and Conditions page.</a>",
    "product_finish_date": None,
    "discount_amount": None,
    "href": "https://localhost:5000/promotions/18336",
    "sale_start_date": "2012-08-14",
    "product_start_date": None,
    "room_codes": None,
    "flags": None,
    "products": PRODUCTS_DATA,
    "promotion_code": "BZ-NOAM10",
    "min_travellers": 0
}



from __future__ import unicode_literals

from ..base import BaseModel, Resource

from .agency import Agency


class AgentRole(BaseModel):
    _as_is_fields = ['id', 'name']


class AgentPhoneNumber(BaseModel):
    _as_is_fields = ['number', 'type']


class Agent(Resource):
    _resource_name = 'agents'
    _is_listable = False

    _as_is_fields = [
        'id',
        'href',
        'first_name',
        'last_name',
        'email',
        'username',
        'active',
    ]
    _model_fields = [
        ('role', AgentRole),
    ]
    _model_collection_fields = [
        ('phone_numbers', AgentPhoneNumber),
    ]
    _resource_fields = [('agency', Agency)]



# Python 2 and 3
from __future__ import unicode_literals

from ..base import Resource


class Document(Resource):
    _resource_name = 'documents'

    _is_listable = False

    _as_is_fields = [
        'id',
        'href',
        'type',
        'audience',
        'content',
        'mime_type',
    ]
    _date_time_fields_utc = ['date_created']
    _resource_fields = [
        ('booking', 'Booking'),
    ]


class Invoice(Resource):
    _resource_name = 'invoices'

    _is_listable = False

    _as_is_fields = [
        'id',
        'href',
        'audience',
    ]

    _date_time_fields_utc = ['date_created']

    _resource_fields = [
        ('document', Document),
        ('booking', 'Booking'),
    ]



from ..base import Resource, BaseModel

from .service import Service


class Requirement(BaseModel):
    _as_is_fields = ['code']


class Override(Resource):
    _resource_name = 'overrides'

    _as_is_fields = [
        'id',
        'href',
    ]

    @property
    def _resource_fields(self):
        from .booking import Booking
        from .customer import Customer
        from .override_reason import OverrideReason

        return [
            ('customer', Customer),
            ('booking', Booking),
            ('reason', OverrideReason),
        ]

    _model_collection_fields = [
        ('services', Service),
    ]

    _model_fields = [
        ('requirement', Requirement),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.resources.base import Resource

from .agency_chain import AgencyChain
from .checkin import Checkin
from .customer import Customer
from .document import Invoice, Document
from .override import Override
from .requirement import RequirementSet
from .service import Service
from .transaction import Payment, Refund


class Booking(Resource):
    _resource_name = 'bookings'
    _is_parent_resource = True

    _as_is_fields = ['id', 'href', 'external_id', 'currency']
    _price_fields = [
        'amount_owing',
        'amount_paid',
        'amount_pending',
        'commission',
        'tax_on_commission',
    ]
    _date_fields = [
        'date_closed', 'date_of_first_travel', 'date_of_last_travel',
        'balance_due_date',
    ]
    _date_time_fields_utc = ['date_created', ]
    _resource_fields = [
        ('agency', 'Agency'),
        ('agency_chain', AgencyChain),
        ('agent', 'Agent'),
        ('associated_agency', 'Agency'),
        ('booking_company', 'BookingCompany')
    ]

    @property
    def _resource_collection_fields(self):
        return [
            ('checkins', Checkin),
            ('customers', Customer),
            ('documents', Document),
            ('invoices', Invoice),
            ('overrides', Override),
            ('payments', Payment),
            ('refunds', Refund),
            ('requirements', RequirementSet),
            ('services', Service),
        ]

    @property
    def _model_collection_fields(self):
        return [
            ('linked_bookings', Booking),
        ]



from gapipy.resources.base import Resource
from gapipy.models import DossierFeatureParent, DossierFeatureChild


class DossierFeature(Resource):
    _resource_name = 'dossier_features'
    _as_is_fields = [
        'id',
        'href',
        'label',
        'code',
        'description',
        'publish_state',
    ]

    _model_fields = [
        ('parent', DossierFeatureParent),
    ]

    _model_collection_fields = [
        ('children', DossierFeatureChild),
    ]



# flake8: NOQA
from .accommodation import Accommodation
from .activity import Activity
from .merchandise import Merchandise
from .departure import Departure
from .departure_component import DepartureComponent
from .image import Image
from .itinerary import (
    Duration,
    Itinerary,
    ItineraryHighlights,
    ItineraryMedia,
    OptionalActivity,
)
from .promotion import Promotion
from .single_supplement import SingleSupplement
from .tour import Tour
from .tour_category import TourCategory
from .tour_dossier import TourDossier
from .transport import Transport



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.resources.base import Resource

from .state import State


class Country(Resource):

    _resource_name = 'countries'
    _is_parent_resource = True

    _as_is_fields = [
        'id',
        'href',
        'name',
    ]

    _resource_fields = [
        ('continent', 'Continent'),
        ('place', 'Place'),
    ]

    _resource_collection_fields = [
        ('states', State),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from .details import DossierDetail, DossierDetailsMixin
from ..base import BaseModel, Resource
from ..tour.image import Image
from ..tour.tour_category import TourCategoryList
from ..tour.video import Video


class AccommodationDossier(Resource, DossierDetailsMixin):
    _resource_name = 'accommodation_dossiers'

    _as_is_fields = [
        'id',
        'href',
        'type',
        'address',
        'costs',
        'emails',
        'features',
        'flags',
        'is_prepaid',
        'name',
        'phone_numbers',
        'property_type',
        'publish_state',
        'rooms',
        'service_code',
        'service_time',
        'show_on_reservation_sheet',
        'suggested_dossiers',  # FIXME: these are typed references to other *_dossiers
        'website',
    ]

    _date_time_fields_local = [
        'date_created',
        'date_last_modified',
    ]

    _resource_fields = [
        ('location', 'Place'),
        ('primary_country', 'Country'),
        ('dossier_segment', 'DossierSegment'),
    ]

    _model_collection_fields = [
        ('categories', TourCategoryList),
        ('details', DossierDetail),
        ('images', Image),
        ('videos', Video),
        ('visited_cities', 'Place'),
        ('visited_countries', 'Country'),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from ..base import Resource
from ..tour.image import Image
from ..tour.video import Video
from .details import DossierDetail


class PlaceDossier(Resource):
    _resource_name = 'place_dossiers'

    _as_is_fields = [
        'id',
        'href',
        'type',
        'flags',
        'name',
        'publish_state',
    ]

    _date_time_fields_local = ['date_created', 'date_last_modified']

    _resource_fields = [
        ('place', 'Place')
    ]

    _model_collection_fields = [
        ('details', DossierDetail),
        ('images', Image),
        ('videos', Video),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models.base import BaseModel
from gapipy.models import AddOn, Address, DepartureRoom, PP2aPrice
from gapipy.resources.booking_company import BookingCompany
from gapipy.resources.product import Product

from .tour_dossier import TourDossier
from .departure_component import DepartureComponent


class LocalPayment(BaseModel):
    _as_is_fields = [
        'amount',
        'currency',
        'label',
    ]


class Departure(Product):

    _resource_name = 'departures'

    _is_listable = True

    _is_parent_resource = True

    _as_is_fields = [
        'id',
        'href',
        'name',
        'availability',
        'flags',
        'nearest_start_airport',
        'nearest_finish_airport',
        'product_line',
        'sku',
        'requirements',
        'program',
    ]
    _date_fields = [
        'start_date',
        'finish_date',
    ]
    _date_time_fields_utc = [
        'date_created',
        'date_last_modified',
        'date_cancelled',
    ]
    _date_time_fields_local = [
        'latest_arrival_time',
        'earliest_departure_time',
    ]
    _resource_fields = [
        ('tour', 'Tour'),
        ('tour_dossier', TourDossier),
    ]
    _resource_collection_fields = [
        ('components', DepartureComponent),
    ]
    _model_fields = [
        ('start_address', Address),
        ('finish_address', Address),
    ]
    _model_collection_fields = [
        ('addons', AddOn),
        ('booking_companies', BookingCompany),
        ('local_payments', LocalPayment),
        ('lowest_pp2a_prices', PP2aPrice),
        ('rooms', DepartureRoom),
        ('structured_itineraries', 'Itinerary'),
    ]



import time
from unittest import TestCase, skipUnless

from gapipy import cache


class SimpleCacheTestCase(TestCase):

    def test_get_set(self):
        c = cache.SimpleCache()
        c.set('foo', {'xyz': 'bar'})
        data = c.get('foo')
        self.assertEqual(data, {'xyz': 'bar'})

    def test_get_set_resource_id(self):
        c = cache.SimpleCache()
        c.set('foo:100', {'id': 100, 'xyz': 'bar'})
        data = c.get('foo:100')
        self.assertEqual(data, {'id': 100, 'xyz': 'bar'})

    def test_delete(self):
        c = cache.SimpleCache()
        c.set('xyz', {'foo': 'bar'})
        self.assertEqual(c.get('xyz'), {'foo': 'bar'})
        c.delete('xyz')
        self.assertEqual(c.get('xyz'), None)


def _redis_installed():
    """Returns True if and only if the redis module is available"""
    try:
        import redis  # NOQA
        return True
    except ImportError:
        return False


@skipUnless(_redis_installed(), 'redis is not installed')
class RedisCacheTestCase(TestCase):
    def make_cache(self):
        return cache.RedisCache(key_prefix='gapi-python-client-test-case:')

    def tearDown(self):
        self.make_cache().clear()

    def test_get_set(self):
        c = self.make_cache()
        c.set('a', {'id': 1})
        self.assertEqual(c.get('a'), {'id': 1})

        c.set('a', {'results': (1, 2, 3)})
        self.assertEqual(c.get('a'), {'results': (1, 2, 3)})

    def test_expire(self):
        c = self.make_cache()
        c.set('expire_me', {0: 0}, timeout=1)
        time.sleep(2)
        self.assertEqual(c.get('expire_me'), None)

    def test_delete(self):
        c = self.make_cache()
        c.set('delete_me', {0: 0})
        self.assertEqual(c.get('delete_me'), {0: 0})
        c.delete('delete_me')
        self.assertEqual(c.get('delete_me'), None)



# Python 2 and 3
from __future__ import unicode_literals

from .base import Resource


class Product(Resource):
    """
    This is used to mark those resources that
    can be attached to a Promotion
    """
    pass



# -*- coding: utf-8 -*-
# Python 2 and 3
from __future__ import unicode_literals

import datetime

from gapipy.models.base import BaseModel
from gapipy.resources.base import Resource
from gapipy.resources.booking_company import BookingCompany
from gapipy.resources.tour.image import Image, MAP_TYPE
from gapipy.utils import (
    DurationLabelMixin,
    LocationLabelMixin,
    duration_label,
    enforce_string_type,
)

class RippleScore(BaseModel):
    _as_is_fields = [
        'score',
        'end_date',
        'start_date',
    ]


class OptionalActivity(BaseModel):
    _resource_fields = [
        ('activity_dossier', 'ActivityDossier'),
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.activity_dossier)


class Duration(BaseModel):
    _as_is_fields = [
        'min_hr', 'max_hr',
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {},{}>'.format(self.__class__.__name__, self.min_hr, self.max_hr)

    @property
    def label(self):
        return duration_label(self.min_hr, self.max_hr)


class ItineraryComponent(BaseModel, DurationLabelMixin, LocationLabelMixin):
    _as_is_fields = [
        'description',
        'distance_km',
        'end_time',
        'instructions',
        'is_overnight',
        'start_time',
        'summary',
        'time_period',
        'type',
    ]
    _resource_fields = [
        ('start_location', 'Place'),
        ('end_location', 'Place'),
        ('accommodation_dossier', 'AccommodationDossier'),
        ('activity_dossier', 'ActivityDossier'),
        ('transport_dossier', 'TransportDossier'),
    ]
    _model_fields = [
        ('duration', 'Duration'),
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.type)

    @property
    def is_free_time(self):
        return self.type == 'FREE_TIME'

    @property
    def is_accommodation(self):
        return self.type == 'ACCOMMODATION'

    @property
    def is_activity(self):
        return self.type == 'ACTIVITY'

    @property
    def is_transport(self):
        return self.type == 'TRANSPORT'


class ItineraryDay(BaseModel):
    _as_is_fields = [
        'id',
        'day',
        'label',
        'summary',
        'description',
        'instructions',
        'meals',
    ]
    _resource_fields = [
        ('start_location', 'Place'),
        ('end_location', 'Place'),
    ]
    _model_collection_fields = [
        ('components', ItineraryComponent),
        ('optional_activities', OptionalActivity),
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.day)


class ValidDuringRange(BaseModel):
    _date_fields = [
        'end_date',
        'start_date',
    ]

    def is_expired(self):
        return not self.is_valid_during_range(datetime.date.today(), None)

    def is_valid_today(self):
        return self.is_valid_on_date(datetime.date.today())

    def is_valid_during_range(self, start_date, end_date):
        if start_date and not end_date:
            return self.is_valid_on_or_after_date(start_date)

        if not start_date and end_date:
            return self.is_valid_on_or_before_date(end_date)

        if not start_date and not end_date:
            return self.is_valid_sometime()

        if start_date == end_date:
            return self.is_valid_on_date(start_date)

        # start_date and end_date are both not None and not equal
        return (
            start_date <= end_date
        ) and (
            self.is_valid_sometime()
        ) and all([
            self.start_date is None or self.start_date <= end_date,
            self.end_date is None or self.end_date >= start_date,
        ])

    def is_valid_on_or_after_date(self, date):
        return (
            self.end_date is None
        ) or (
            self.start_date is None and
            self.end_date >= date
        ) or (
            self.start_date is not None and
            self.start_date <= self.end_date and
            self.end_date >= date
        )

    def is_valid_on_or_before_date(self, date):
        return (
            self.start_date is None
        ) or (
            self.end_date is None and
            self.start_date <= date
        ) or (
            self.end_date is not None and
            self.start_date <= self.end_date and
            self.start_date <= date
        )

    def is_valid_on_date(self, date):
        return all([
            self.start_date is None or self.start_date <= date,
            self.end_date is None or self.end_date >= date,
        ])

    def is_valid_sometime(self):
        return (
            self.start_date is None or
            self.end_date is None or
            self.start_date <= self.end_date
        )

    @enforce_string_type
    def __repr__(self):
        return '<{} ({} - {})>'.format(self.__class__.__name__, self.start_date, self.end_date)


class DetailType(BaseModel):
    _as_is_fields = [
        'id',
        'code',
        'name',
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.code)


class Detail(BaseModel):
    _as_is_fields = ['body']

    _model_fields = [
        ('type', DetailType),
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{} {}: {}>'.format(self.__class__.__name__, self.type.code, self.body[:100])


class ItineraryMedia(Resource):
    _resource_name = 'itinerary_media'

    _as_is_fields = [
        'id',
        'type',
        'image',
        'video_thumb',
        'videos',
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.type)


class ItineraryHighlights(Resource):
    _resource_name = 'itinerary_highlights'

    _as_is_fields = [
        'id',
        'name',
        'description',
        'media',
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)


class Itinerary(Resource):
    _resource_name = 'itineraries'

    _is_parent_resource = True

    _as_is_fields = [
        'id',
        'href',
        'name',
        'flags',
        'duration',
        'meals_included',
        'meals_budget',
        'packing_lists',
        'publish_state',
        'site_links',
        'variation_id',
    ]
    _resource_fields = [
        ('end_location', 'Place'),
        ('start_location', 'Place'),
        ('tour_dossier', 'TourDossier'),
    ]
    _model_collection_fields = [
        ('booking_companies', BookingCompany),
        ('days', ItineraryDay),
        ('details', Detail),
        ('images', Image),
        ('valid_during_ranges', ValidDuringRange),
        ('variations', 'Itinerary'),
    ]

    _model_fields = [
        ('ripple_score', RippleScore),
    ]

    _resource_collection_fields = [
        ('media', ItineraryMedia),
        ('highlights', ItineraryHighlights),
    ]

    def get_map_image(self):
        """
        Returns the first Image in our list that claims to be a map. Returns
        None if no Images are listed, or if none are marked as a map image.
        """
        for image in self.images:
            if getattr(image, 'type', None) == MAP_TYPE:
                return image
        return None

    def get_map_url(self):
        """
        Returns the URL of this itinerary's map image (or None if no map image
        is found, see get_map_image())
        """
        image = self.get_map_image()
        if not image or not getattr(image, 'file', None):
            return None
        return image.file.url

    def is_expired(self):
        return all([
            vrange.is_expired()
            for vrange in self.valid_during_ranges
        ])

    def is_valid_today(self):
        return any([
            vrange.is_valid_today()
            for vrange in self.valid_during_ranges
        ])

    def is_valid_during_range(self, start_date, end_date):
        return any([
            vrange.is_valid_during_range(start_date, end_date)
            for vrange in self.valid_during_ranges
        ])

    def is_valid_on_or_after_date(self, date):
        return any([
            vrange.is_valid_on_or_after_date(date)
            for vrange in self.valid_during_ranges
        ])

    def is_valid_on_or_before_date(self, date):
        return any([
            vrange.is_valid_on_or_before_date(date)
            for vrange in self.valid_during_ranges
        ])

    def is_valid_on_date(self, date):
        return any([
            vrange.is_valid_on_date(date)
            for vrange in self.valid_during_ranges
        ])

    def is_valid_sometime(self):
        return any([
            vrange.is_valid_sometime()
            for vrange in self.valid_during_ranges
        ])



from functools import wraps
from itertools import islice

from requests import HTTPError
from requests.status_codes import codes

from .request import APIRequestor


HTTPERRORS_MAPPED_TO_NONE = (
    codes.FORBIDDEN,  # 403
    codes.NOT_FOUND,  # 404
    codes.GONE,  # 410
)


def _check_listable(func):

    @wraps(func)
    def inner(query, *args, **kwargs):
        if not (query.resource._is_listable or query.parent):
            raise ValueError(
                'The {0} resource is not listable and/or is only available as a subresource'.format(
                    query.resource.__name__))

        return func(query, *args, **kwargs)

    return inner


class Query(object):

    def __init__(self, client, resource, filters=None, parent=None, raw_data=None):
        self._client = client
        self._raw_data = raw_data or {}
        self.resource = resource
        self._filters = filters or {}
        self.parent = parent

    def _to_dict(self):
        # Used by Resource when converting nested Query objects into
        # serializable types.
        return self._raw_data

    def options(self):
        return self.resource.options(client=self._client)

    def get(self, resource_id, variation_id=None, cached=True, headers=None):
        """
        Returns an instance of the query resource with the given `resource_id`
        (and optional `variation_id`) or `None` if the resource with the given
        id does not exist.

        If the resource is available in the cache and `cached` is True, no http
        request will be made. If `cached` is False, a fresh request will be
        made for the resource, which will be re-added to the cache. This is a
        good method to invalidate persistent cache backend after receiving a
        webhook that a resource has changed.
        """
        key = self.query_key(resource_id, variation_id)

        try:
            data = self.get_resource_data(
                resource_id,
                variation_id=variation_id,
                cached=cached,
                headers=headers
            )
        except HTTPError as e:
            if e.response.status_code in HTTPERRORS_MAPPED_TO_NONE:
                return None
            raise e
        resource_object = self.resource(data, client=self._client)
        return resource_object

    def get_resource_data(self, resource_id, variation_id=None, cached=True, headers=None):
        '''
        Returns a dictionary of resource data, which is used to initialize
        a Resource object in the `get` method.
        '''
        key = self.query_key(resource_id, variation_id)
        resource_data = None
        if cached:
            resource_data = self._client._cache.get(key)
            if resource_data is not None:
                return resource_data

        # Cache miss; get fresh data from the backend, set in cache
        requestor = APIRequestor(self._client, self.resource)
        out = requestor.get(resource_id, variation_id=variation_id, headers=headers)
        if out is not None:
            self._client._cache.set(key, out)
        self._filters = {}
        return out

    def purge_cached(self, resource_id, variation_id=None):
        key = self.query_key(resource_id, variation_id)
        return self._client._cache.delete(key)

    def is_cached(self, resource_id, variation_id=None):
        key = self.query_key(resource_id, variation_id)
        return self._client._cache.is_cached(key)

    def query_key(self, resource_id=None, variation_id=None):
        """Returns a unique key for the information used to fetch the resource(s)
        in this query. Currently used for creating cache keys.
        """
        if not resource_id:
            return self.resource._resource_name

        parts = [self.resource._resource_name, str(resource_id)]
        if variation_id:
            parts.append(str(variation_id))
        if self._client.api_language:
            parts.append(self._client.api_language)

        if self._client.application_key:
            part = self._client.application_key.split('_')[0]
            if part == self._client.application_key or part.strip(' ') != 'test':
                return ':'.join(parts)
            parts.append(part)
        return ':'.join(parts)

    @_check_listable
    def all(self, limit=None):
        """Generator of instances of the query resource. If limit is set to a
        positive integer `n`, only return the first `n` results.
        """

        requestor = APIRequestor(
            self._client,
            self.resource,
            params=self._filters,
            parent=self.parent
        )
        # use href when available; this change should be transparent
        # introduced: 2.20.0
        href =  None
        if isinstance(self._raw_data, dict):
            href = self._raw_data.get('href')
        # generator to fetch list resources
        generator = requestor.list(href)
        # reset filters in case they were set on this query
        self._filters = {}

        if limit:
            if isinstance(limit, int) and limit > 0:
                generator = islice(generator, limit)
            else:
                raise ValueError('`limit` must be a positive integer')

        for result in generator:
            yield self.resource(result, client=self._client, stub=True)

    def filter(self, **kwargs):
        """Add filter arguments to the query.

        For example, if `query` is a Query for the TourDossier ressource, then
        `query.filter(name='Amazing Adventure')` will return a query containing
        only dossiers whose names contain 'Amazing Adventure'.
        """
        self._filters.update(kwargs)
        return self

    @_check_listable
    def count(self):
        """Returns the number of element in the query."""

        requestor = APIRequestor(
            self._client,
            self.resource,
            params=self._filters,
            parent=self.parent
        )
        response = requestor.list_raw()
        out = response.get('count')
        self._filters = {}
        return out

    def create(self, data_dict, headers=None):
        """Create an instance of the query resource using the given data"""
        return self.resource.create(self._client, data_dict, headers=headers)

    def first(self):
        """
        Returns the first object of a query, returns None if no match is found.
        """
        return next(self.all(), None)

    def __iter__(self):
        """Provided as a convenience so that Query objects can be iterated
        without calling `all`.

            i.e.  `for dossier in dossiers.filter(name='Peru')`
                instead of `for dossier in dossiers.filter(name='Peru').all()`
        """
        return self.all()



# Python 2 and 3
from __future__ import unicode_literals

from future.utils import with_metaclass

from gapipy.models import (
    ArrivalFlightDetail,
    AssociatedService,
    DepartureFlightDetail,
    DepartureServiceRoom,
    DocumentInfo,
    IncompleteRequirement,
    InternationalTicketNumber,
    TravellerHeight,
)
from gapipy.models.base import BaseModel
from gapipy.resources.base import Resource
from gapipy.resources.tour import (
    Departure,
    DepartureComponent,
    Accommodation,
    Activity,
    Merchandise,
    Transport,
    SingleSupplement,
)
from gapipy.utils import get_resource_class_from_resource_name

from .customer import Customer
from .declined_reason import DeclinedReason


class TypeBasedServiceMeta(type):
    def __call__(cls, *args, **kwargs):
        # Attempt to extract the type from the data. If it's not found for
        # whatever reason, we can't assume what the user was attempting, and
        # we'll simply ignore any metaclass magic.
        resource_name = args[0].get('type', None) if args else None

        if resource_name is None:
            return type.__call__(cls, *args, **kwargs)

        new_class = get_resource_class_from_resource_name(resource_name)
        return type.__call__(new_class, *args, **kwargs)


# with_metaclass is Python 2 and Python 3 method to allow metaclasses
class Service(with_metaclass(TypeBasedServiceMeta, Resource)):
    _resource_name = 'services'
    _is_listable = False
    _is_parent_resource = True

    @property
    def _as_is_fields(self):
        return [
            'id',
            'href',
            'name',
            'type',
            'sub_type',
            'applied_promotion',
            'flags',
            'status',
            'status_transitions',
        ]

    @property
    def _date_fields(self):
        return [
            'start_date',
            'finish_date',
        ]

    @property
    def _date_time_fields_utc(self):
        return [
            'date_created',
            'date_confirmed',
            'date_cancelled',
            'option_expiry_date',
        ]

    @property
    def _date_time_fields_local(self):
        return []

    @property
    def _price_fields(self):
        return [
            'commission',
            'purchase_price',
        ]

    @property
    def _resource_fields(self):
        return [
            ('booking', 'Booking'),
            ('declined_reason', DeclinedReason),
        ]

    @property
    def _model_collection_fields(self):
        return [
            ('customers', Customer),
            ('documents', DocumentInfo),
        ]


class ServiceProduct(Service):
    def _get_product(self):
        # Return the appropriate product based on the parent service.
        mapping = {
            'departure_services': Departure,
            'accommodation_services': Accommodation,
            'transport_services': Transport,
            'activity_services': Activity,
            'merchandise_services': Merchandise,
            'single_supplement_services': SingleSupplement,
        }
        return mapping[self._resource_name]

    @property
    def _resource_fields(self):
        return super(ServiceProduct, self)._resource_fields + [
            ('product', self._get_product()),
        ]


class Airport(BaseModel):
    _as_is_fields = ['code', 'name']


class Airline(BaseModel):
    _as_is_fields = ['code', 'name']


class FlightServiceSegment(Resource):
    _as_is_fields = [
        'id',
        'href',
        'flight_number',
        'booking_class',
        'technical_stops',
    ]

    _model_fields = [
        ('airline', Airline),
        ('origin_airport', Airport),
        ('destination_airport', Airport),
    ]

    _date_time_fields_local = [
        'departure_date',
        'arrival_date',
    ]

    _resource_fields = [
        ('flight_service', 'FlightService'),
    ]


class DepartureServiceFlight(BaseModel):
    _model_collection_fields = [
        ('flights', FlightServiceSegment)
    ]

    _resource_fields = [
        ('customer', Customer),
    ]


class DepartureService(ServiceProduct):
    _resource_name = 'departure_services'

    @property
    def _as_is_fields(self):
        return super(DepartureService, self)._as_is_fields + [
            'requirements',
        ]

    @property
    def _price_fields(self):
        return super(DepartureService, self)._price_fields + [
            'deposit',
        ]

    @property
    def _resource_fields(self):
        return (super(DepartureService, self)._resource_fields + [
            ('itinerary', 'Itinerary'),
            ('original_departure_service', DepartureService),
        ])

    @property
    def _resource_collection_fields(self):
        return super(DepartureService, self)._resource_collection_fields + [
            ('components', DepartureComponent),
        ]

    @property
    def _model_collection_fields(self):
        return super(DepartureService, self)._model_collection_fields + [
            ('arrival_flight_details', ArrivalFlightDetail),
            ('departure_flight_details', DepartureFlightDetail),
            ('incomplete_requirements', IncompleteRequirement),
            ('international_ticket_numbers', InternationalTicketNumber),
            ('rooms', DepartureServiceRoom),
            ('arriving_flights', DepartureServiceFlight),
            ('departing_flights', DepartureServiceFlight),
            ('traveller_heights', TravellerHeight),
        ]


class AccommodationService(ServiceProduct):
    _resource_name = 'accommodation_services'

    @property
    def _as_is_fields(self):
        return super(AccommodationService, self)._as_is_fields + [
            'room',
        ]

    @property
    def _model_collection_fields(self):
        return super(AccommodationService, self)._model_collection_fields + [
            ('associated_services', AssociatedService),
        ]


class TransportService(ServiceProduct):
    _resource_name = 'transport_services'

    @property
    def _as_is_fields(self):
        return super(TransportService, self)._as_is_fields + [
            'flight_number', 'pickup_time',
        ]

    @property
    def _model_collection_fields(self):
        return super(TransportService, self)._model_collection_fields + [
            ('associated_services', AssociatedService),
        ]


class ActivityService(ServiceProduct):
    _resource_name = 'activity_services'

    @property
    def _model_collection_fields(self):
        return (super(ActivityService, self)._model_collection_fields + [
            ('arrival_flight_details', ArrivalFlightDetail),
            ('associated_services', AssociatedService),
            ('incomplete_requirements', IncompleteRequirement),
        ])


class MerchandiseService(ServiceProduct):
    _resource_name = 'merchandise_services'

    @property
    def _model_collection_fields(self):
        return (super(MerchandiseService, self)._model_collection_fields + [
            ('associated_services', AssociatedService),
        ])


class SingleSupplementService(ServiceProduct):
    _resource_name = 'single_supplement_services'

    @property
    def _model_collection_fields(self):
        return super(SingleSupplementService, self)._model_collection_fields + [
            ('associated_services', AssociatedService),
        ]


class InsuranceService(Service):
    _resource_name = 'insurance_services'

    @property
    def _as_is_fields(self):
        return super(InsuranceService, self)._as_is_fields + [
            'policy_details_url',
            'policy_number',
            'policy_emergency_phone_number',
            'policy_provider',
            'policy_type',
        ]

    @property
    def _price_fields(self):
        return super(InsuranceService, self)._price_fields + [
            'policy_amount',
            'insurable_amount',
        ]


class FeeService(Service):
    _resource_name = 'fee_services'

    @property
    def _as_is_fields(self):
        return super(FeeService, self)._as_is_fields + [
            'description',
            'related_service',
        ]


class FlightStatus(BaseModel):
    """
    Represent a flight_status attached to the FlightService
    """
    _as_is_fields = ['id', 'href']


class FlightService(Service):
    _resource_name = 'flight_services'

    @property
    def _as_is_fields(self):
        return super(FlightService, self)._as_is_fields + [
            'record_locator',
            'itinerary_url',
            'segments',
        ]

    @property
    def _price_fields(self):
        price_fields = super(FlightService, self)._price_fields

        return [
            field for field in price_fields
            if field not in ('commission', )
        ]

    @property
    def _date_time_fields_utc(self):
        date_fields = super(FlightService, self)._date_time_fields_utc

        return [
            field for field in date_fields
        ]

    @property
    def _model_fields(self):
        return super(FlightService, self)._model_fields + [
            ('flight_status', FlightStatus)
        ]

    @property
    def _model_collection_fields(self):
        return super(FlightService, self)._model_collection_fields + [
            ('associated_services', AssociatedService),
        ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models import AdvertisedDeparture
from gapipy.query import Query
from gapipy.resources.base import Resource
from gapipy.resources.booking_company import BookingCompany
from gapipy.utils import get_resource_class_from_class_name


MAP_IMAGE_TYPE = 'MAP'
BANNER_IMAGE_TYPE = 'BANNER'


class TourDossier(Resource):

    _resource_name = 'tour_dossiers'

    _as_is_fields = [
        'id',
        'href',
        'categories',
        'description',
        'details',
        'product_line',
        'geography',
        'images',
        'name',
        'site_links',
        'slug',
    ]
    _date_fields = [
        'departures_start_date',
        'departures_end_date',
    ]
    _resource_fields = [
        ('tour', 'Tour'),
    ]
    _resource_collection_fields = [
        ('departures', 'Departure'),
    ]
    _model_collection_fields = [
        ('advertised_departures', AdvertisedDeparture),
        ('booking_companies', BookingCompany),
        ('structured_itineraries', 'Itinerary'),
    ]

    def _set_resource_collection_field(self, field, value):
        """
        Overridden to ensure that the `departures` query has the right
        parent resource (i.e. the tour and not the tour dossier).
        """
        if field == 'departures':
            resource_cls = get_resource_class_from_class_name('Departure')
            # Tour dossiers always have the same id as the corresponding tour
            parent = ('tours', self.id, None)
            query = Query(self._client, resource_cls, parent=parent, raw_data=value)
            setattr(self, field, query)
        else:
            return super(TourDossier, self)._set_resource_collection_field(field, value)

    def _get_image_url(self, image_type):
        for image in self.images:
            if image['type'] == image_type:
                return image['image_href']

    def get_map_url(self):
        return self._get_image_url(MAP_IMAGE_TYPE)

    def get_banner_url(self):
        return self._get_image_url(BANNER_IMAGE_TYPE)

    def get_visited_countries(self):
        return [country['name'] for country in self.geography['visited_countries']]

    def get_trip_detail(self, label):
        for detail in self.details:
            if detail['detail_type']['label'] == label:
                return detail['body']

    def get_category_name(self, label):
        for category in self.categories:
            if category['category_type']['label'] == label:
                return category['name']



from gapipy.utils import get_resource_class_from_class_name

# get the Promotion class
#
# NOTE: this is intentional, and needed to ensure we
#       are loading the `correct` Promotion class
Promotion  = get_resource_class_from_class_name('Promotion')


class PricePromotion(Promotion):
    """
    This extends the definition of the Promotion class
    adding the ``amount`` price field, for use in the
    `Prices` model
    """
    @property
    def _price_fields(self):
        return super(PricePromotion, self)._price_fields + ['amount']



# Python 2 and 3
from __future__ import unicode_literals

from ..base import Resource
from ...models import DossierFeatureParent, DossierFeatureChild


class DossierSegment(Resource):
    _resource_name = 'dossier_segments'
    _as_is_fields = ['id', 'href', 'label', 'code', 'description']

    _model_fields = [
        ('parent', DossierFeatureParent),
    ]

    _model_collection_fields = [
        ('children', DossierFeatureChild),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models import Address, AgencyDocument
from gapipy.models.base import BaseModel
from gapipy.resources.base import Resource
from gapipy.resources.booking_company import BookingCompany
from gapipy.resources.tour import Promotion

from .agency_chain import AgencyChain


class AgencyEmail(BaseModel):
    _as_is_fields = ['type', 'address']


class Agency(Resource):
    _resource_name = 'agencies'
    _is_listable = False
    _is_parent_resource = True

    _as_is_fields = [
        'id',
        'href',
        'name',
        'booking_currencies',
        'latitude',
        'longitude',
        'transactional_email',
        'communication_preferences',
        'override_agency_secondary',
        'passenger_notifications',
        'agent_notifications',
        'preferred_display_name',
    ]
    _date_time_fields_local = [
        'date_created',
    ]
    _model_fields = [
        ('address', Address),
    ]
    _resource_fields = [
        ('agency_chain', AgencyChain),
    ]
    _model_collection_fields = [
        ('agency_chains', 'AgencyChain'),
        ('booking_companies', BookingCompany),
        ('documents', AgencyDocument),
        ('emails', AgencyEmail),
    ]
    _resource_collection_fields = [
        ('agents', 'Agent'),
        ('promotions', Promotion),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.resources.base import Resource
from gapipy.resources.booking_company import BookingCompany


class AgencyChain(Resource):
    _resource_name = 'agency_chains'

    _as_is_fields = [
        'id',
        'href',
        'name',
        'agencies',
        'agent_notifications',
        'communication_preferences',
        'flags',
        'payment_options',
        'passenger_notifications',
    ]

    _date_time_fields_local = [
        'date_created',
        'date_last_modified',
    ]

    _resource_fields = [
        ('booking_company', BookingCompany),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.utils import get_resource_class_from_resource_name
from gapipy.resources.base import Resource
from gapipy.resources.product import Product


class PromotionProduct(Resource):
    """
    The `products` referenced in a Promotion object are not valid resources (due
    to the existence of `type`) so they must be wrapped in this model.

    When the resource is fixed, the hacks here can simply be replaced with
    `gapipy.models.base.RelatedResourceMixin`
    """
    def __init__(self, data, **kwargs):
        # Fetch the resource class using the `type`, and then derive field
        # attributes from that class.
        klass = get_resource_class_from_resource_name(data['type'])
        # Python 2 and 3
        # inefficient on Python 2 to list items()
        for k, v in list(klass.__dict__.items()):
            if 'fields' in k and isinstance(v, list):
                setattr(self, k, getattr(klass, k))

        self._resource_name = data['type']
        self._as_is_fields = self._as_is_fields + ['type', 'sub_type']

        super(PromotionProduct, self).__init__(data, **kwargs)


class Promotion(Resource):

    _resource_name = 'promotions'

    _as_is_fields = [
        'id',
        'href',
        'commission_rate',
        'currencies',
        'flags',
        'min_travellers',
        'name',
        'promotion_code',
        'room_codes',
        'terms_and_conditions',
    ]

    _date_time_fields_utc = [
        'date_created',
        'date_last_modified',
        'sale_finish_datetime',
        'sale_start_datetime',
    ]

    _date_fields = [
        'product_finish_date',
        'product_start_date',
        'sale_finish_date',
        'sale_start_date',
    ]
    _price_fields = [
        'discount_amount',
        'discount_percent',
    ]

    _model_collection_fields = [
        ('products', Product),
    ]

    def _set_model_collection_field(self, field, data):
        # Only products needs special treatment.
        if field != 'products':
            return super(Promotion, self)._set_model_collection_field(field, data)

        product_stubs = []
        # Each product can be a different resource, so derive the resource from
        # the "type" within the stubbed object.
        for product in data:
            stub = PromotionProduct(product, client=self._client, stub=True)
            product_stubs.append(stub)
        setattr(self, field, product_stubs)



import json
import sys
import unittest

from mock import MagicMock, patch
from requests import HTTPError, Response

from gapipy.client import Client
from gapipy.query import Query
from gapipy.resources import Activity, Departure, Tour, TourDossier
from gapipy.resources.base import Resource
from gapipy.utils import get_available_resource_classes

from .fixtures import (
    PPP_TOUR_DATA, TOUR_DOSSIER_LIST_DATA, DUMMY_DEPARTURE, DUMMY_PROMOTION,
)


class QueryKeyTestCase(unittest.TestCase):

    def setUp(self):
        # Any ol' resource will do.
        self.client = Client(application_key='test_abcd')
        self.resource = get_available_resource_classes()[0]
        self.resource_name = self.resource._resource_name

    def test_query_key_with_env_test(self):
        client = Client(application_key='test_abcd')
        output = client.tour_dossiers.query_key(123)
        expected = 'tour_dossiers:123:test'
        self.assertEqual(output, expected)

    def test_query_key_test(self):
        client = Client(application_key='test')
        output = client.tour_dossiers.query_key(123)
        expected = 'tour_dossiers:123'
        self.assertEqual(output, expected)

    def test_query_key_test_with_no_underscore_with_test(self):
        client = Client(application_key='testringnounderscore')
        output = client.tour_dossiers.query_key(123)
        expected = 'tour_dossiers:123'
        self.assertEqual(output, expected)

    def test_query_key_with_no_underscore(self):
        client = Client(application_key='somestringnounderscore')
        output = client.tour_dossiers.query_key(123)
        expected = 'tour_dossiers:123'
        self.assertEqual(output, expected)

    def test_query_key_with_underscore_live(self):
        client = Client(application_key='live_abs')
        output = client.tour_dossiers.query_key(123)
        expected = 'tour_dossiers:123'
        self.assertEqual(output, expected)

    def test_query_key_with_env_with_language(self):
        self.client = Client(application_key='test_abcd')
        self.client.api_language = 'de'
        output = self.client.tour_dossiers.query_key(123)
        expected = 'tour_dossiers:123:de:test'
        self.assertEqual(output, expected)

    def test_query_key_with_env_with_both_ids(self):
        client = Client(application_key='test_abcd')
        output = client.tour_dossiers.query_key(1, 2)
        expected = 'tour_dossiers:1:2:test'
        self.assertEqual(output, expected)

    def test_query_key_with_env_with_one_id(self):
        client = Client(application_key='test_abcd')
        output = client.tour_dossiers.query_key(1)
        expected = 'tour_dossiers:1:test'
        self.assertEqual(output, expected)

    def test_query_key_with_language(self):
        self.client = Client(application_key='')
        self.client.api_language = 'de'
        query = Query(self.client, self.resource)
        key = query.query_key(1)
        expected = '{}:1:de'.format(self.resource_name)
        self.assertEqual(key, expected)

        # Unsetting the language should also remove it from the key
        self.client.api_language = None
        key = query.query_key(1)
        expected = '{}:1'.format(self.resource_name)
        self.assertEqual(key, expected)

    def test_query_key_with_variation_id(self):
        query = Query(self.client, self.resource)
        key = query.query_key(1, 2)
        expected = '{}:1:2:test'.format(self.resource_name)
        self.assertEqual(key, expected)

    def test_query_key_no_resource_id(self):
        query = Query(self.client, self.resource)
        key = query.query_key()
        self.assertEqual(key, self.resource_name)


class QueryTestCase(unittest.TestCase):

    def setUp(self):
        self.client = Client()
        self.cache = self.client._cache
        self.cache.clear()

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_TOUR_DATA)
    def test_get_instance_by_id(self, mock_request):
        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsInstance(t, Tour)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_instance_with_forbidden_id(self, mock_request):
        response = Response()
        response.status_code = 403
        http_error = HTTPError(response=response)
        mock_request.side_effect = http_error

        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsNone(t)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_instance_with_non_existing_id(self, mock_request):
        response = Response()
        response.status_code = 404
        http_error = HTTPError(response=response)
        mock_request.side_effect = http_error

        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsNone(t)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_instance_with_gone_id(self, mock_request):
        response = Response()
        response.status_code = 410
        http_error = HTTPError(response=response)
        mock_request.side_effect = http_error

        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsNone(t)

    @patch('gapipy.request.APIRequestor._request')
    def test_get_instance_by_id_with_non_404_error(self, mock_request):
        response = Response()
        response.status_code = 401
        http_error = HTTPError(response=response)
        mock_request.side_effect = http_error

        query = Query(self.client, Tour)
        with self.assertRaises(HTTPError) as cm:
            query.get(1234)

        self.assertEqual(cm.exception.response.status_code, 401)

    @patch('gapipy.request.APIRequestor._request')
    def test_filtered_query(self, mock_request):
        """
        Arguments passed to .filter() are stored on the Query instance but are
        cleared when that query is evaluated.
        """
        # Create a basic filter query for PPP...
        query = Query(self.client, Tour).filter(tour_dossier_code='PPP')
        self.assertEqual(len(query._filters), 1)

        # ... then chain on another filter argument...
        query = query.filter(order_by__asc='departures_start_date')
        self.assertEqual(len(query._filters), 2)

        # ... and force query evaluation, before checking...
        list(query)

        # ... our request was made with the appropriate query args, and...
        mock_request.assert_called_once_with(
            '/tours', 'GET', params={
                'tour_dossier_code': 'PPP',
                'order_by__asc': 'departures_start_date',
            })
        mock_request.reset_mock()

        # ... our stored filter args got reset.
        self.assertEqual(len(query._filters), 0)

        # Check .count() also clears stored filter args appropriately:
        query.filter(
            tour_dossier_code='PPP',
            order_by__desc='departures_start_date').count()
        mock_request.assert_called_once_with(
            '/tours', 'GET', params={
                'tour_dossier_code': 'PPP',
                'order_by__desc': 'departures_start_date',
            })
        mock_request.reset_mock()
        self.assertEqual(len(query._filters), 0)


    @patch('gapipy.request.APIRequestor._request')
    def test_query_reset_filter(self, mock_request):
        query = Query(self.client, Tour)
        query.filter(tour_dossier_code='PPP').count()
        self.assertEqual(query._filters, {})

    def test_listing_non_listable_resource_fails(self):
        message = 'The Activity resource is not listable and/or is only available as a subresource'
        if sys.version_info.major < 3:
            with self.assertRaisesRegexp(ValueError, message):
                Query(self.client, Activity).all()
            with self.assertRaisesRegexp(ValueError, message):
                Query(self.client, Activity).count()
        else:
            with self.assertRaisesRegex(ValueError, message):
                Query(self.client, Activity).all()
            with self.assertRaisesRegex(ValueError, message):
                Query(self.client, Activity).count()

    @patch('gapipy.request.APIRequestor._request', return_value=DUMMY_PROMOTION)
    def test_can_retrieve_single_non_listable_resource(self, mock_request):
        Query(self.client, Activity).get(1234)
        mock_request.assert_called_once_with(
            '/activities/1234', 'GET', additional_headers=None)

    @patch('gapipy.request.APIRequestor._request', return_value=DUMMY_DEPARTURE)
    def test_can_retrieve_single_subresource_without_parent(self, mock_request):
        Query(self.client, Departure).get(1234)
        mock_request.assert_called_once_with(
            '/departures/1234', 'GET', additional_headers=None)

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_count(self, mock_request):
        query = Query(self.client, TourDossier)
        count = query.count()
        self.assertIsInstance(count, int)
        self.assertEqual(count, 3)

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_fetch_all(self, mock_request):

        query = Query(self.client, TourDossier).all()
        self.assertFalse(mock_request.called)

        dossiers = list(query)
        self.assertTrue(mock_request.called)

        self.assertEqual(len(dossiers), 3)
        for dossier in dossiers:
            self.assertIsInstance(dossier, TourDossier)

        mock_request.assert_called_once_with(
            '/tour_dossiers', 'GET', params={})

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_fetch_all_with_limit(self, mock_request):

        query = Query(self.client, TourDossier).all(limit=2)
        dossiers = list(query)

        self.assertEqual(len(dossiers), 2)
        for dossier in dossiers:
            self.assertIsInstance(dossier, TourDossier)

        mock_request.assert_called_once_with(
            '/tour_dossiers', 'GET', params={})

    def test_fetch_all_with_wrong_argument_for_limit(self):
        message = '`limit` must be a positive integer'
        if sys.version_info.major < 3:
            with self.assertRaisesRegexp(ValueError, message):
                query = Query(self.client, Tour).all(limit=-1)
                list(query)  # force the query to evaluate
        else:
            with self.assertRaisesRegex(ValueError, message):
                query = Query(self.client, Tour).all(limit=-1)
                list(query)  # force the query to evaluate


class QueryCacheTestCase(unittest.TestCase):
    def setUp(self):
        self.client = Client(cache_backend='gapipy.cache.SimpleCache')
        self.cache = self.client._cache
        self.cache.clear()

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_TOUR_DATA)
    def test_resources_are_cached(self, mock_request):
        query = Query(self.client, Tour)

        self.assertEqual(self.cache.count(), 0)

        query.get(21346)
        self.assertEqual(self.cache.count(), 1)
        self.assertEqual(len(mock_request.mock_calls), 1)

        query.get(21346)
        self.assertEqual(len(mock_request.mock_calls), 1)

    def test_cached_get_does_not_set(self):
        """
        Regression test https://github.com/gadventures/gapipy/issues/65

        We discovered that when getting a resource, even if it is a hit in our
        local cache we would set the data back into our cache every time. When
        using a cache with a TTL on keys (e.g. Redis) this has the effect of
        resetting the TTL each time that that key is retrieved. This is not the
        expected behaviour wrt cache key TTLs.
        """
        query = Query(self.client, Tour)

        # act like we already have the data in our cache
        mock_cache_get = MagicMock(return_value=PPP_TOUR_DATA)
        self.cache.get = mock_cache_get

        mock_cache_set = MagicMock()
        self.cache.set = mock_cache_set

        query.get(21346)
        self.assertEqual(len(mock_cache_get.mock_calls), 1)
        self.assertEqual(len(mock_cache_set.mock_calls), 0)


class MockResource(Resource):
    _as_is_fields = ['id', 'first_name', 'last_name']
    _resource_name = 'mocks'


@patch('gapipy.request.APIRequestor._request')
class UpdateCreateResourceTestCase(unittest.TestCase):

    def setUp(self):
        self.client = Client()

    def test_object_accessor(self, mock_request):
        data = {'first_name': 'Jon', 'last_name': 'Ive', 'id': None}
        r = MockResource(data, client=self.client)
        self.assertEqual(r.first_name, data['first_name'])

    def test_object_attr_modify(self, mock_request):
        data = {'first_name': 'Jon', 'last_name': 'Ive', 'id': None}
        r = MockResource(data, client=self.client)
        r.first_name = 'Jonathan'
        self.assertEqual(r.first_name, 'Jonathan')

    def test_create_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': None,
        }
        # modify mock_request to return a value
        # this is needed for save -> _fill_fields
        mock_request.return_value = data

        r = MockResource(data, client=self.client)
        r.save()
        mock_request.assert_called_once_with(
            '/mocks', 'POST', data=r.to_json(), additional_headers=None)

    def test_update_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': 1,
        }
        # modify mock_request to return a value
        # this is needed for save -> _fill_fields
        mock_request.return_value = {
            'id': 1,
            'first_name': 'Jonathan',
            'last_name': 'Ive',
        }

        r = MockResource(data, client=self.client)
        r.first_name = 'Jonathan'
        r.save()

        mock_request.assert_called_once_with(
            '/mocks/1', 'PUT', data=r.to_json())

    def test_partial_update_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': 1,
        }
        mock_request.return_value = data
        r = MockResource(data, client=self.client)

        r.first_name = 'Jonathan'
        r_data = {
            'first_name': 'Jonathan',
            'last_name': 'Ive',
            'id': 1
        }
        mock_request.return_value = r_data
        r.save(partial=True)
        changed = {'first_name': 'Jonathan'}
        mock_request.assert_called_once_with(
            '/mocks/1', 'PATCH', data=json.dumps(changed))
        self.assertEqual(r.to_dict(), r_data)

        r.last_name = 'Ivey'
        r_data = {
            'first_name': 'Jonathan',
            'last_name': 'Ivey',
            'id': 1,
        }
        mock_request.return_value = r_data
        r.save(partial=True)
        changed = {'last_name': 'Ivey'}
        mock_request.assert_called_with(
            '/mocks/1', 'PATCH', data=json.dumps(changed))
        self.assertEqual(r.to_dict(), r_data)



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models import Address
from gapipy.models.base import BaseModel
from gapipy.resources.base import Resource


class MedicalDetail(BaseModel):
    _as_is_fields = ['type', 'value']


class MembershipProgram(BaseModel):
    _as_is_fields = ['code', 'label', 'value']


class Customer(Resource):
    _resource_name = 'customers'
    _is_listable = False
    _is_parent_resource = True

    _as_is_fields = [
        'account_email',
        'emergency_contacts',
        'gender',
        'href',
        'id',
        'meal_notes',
        'meal_preference',
        'medical_notes',
        'name',
        'nationality',
        'passport',
        'phone_numbers',
        'place_of_birth',
    ]

    _date_fields = ['date_of_birth', ]

    _model_fields = [
        ('address', Address),
    ]

    _model_collection_fields = [
        ('medical_details', MedicalDetail),
        ('membership_programs', MembershipProgram),
    ]

    @property
    def _resource_collection_fields(self):
        from .booking import Booking
        return [
            ('bookings', Booking),
        ]



#!/usr/bin/env python
# Python 2 and 3
from __future__ import unicode_literals

from ..base import Resource
from ...utils import enforce_string_type


class OverrideReason(Resource):

    _resource_name = 'override_reasons'

    _as_is_fields = ['id', 'href', 'label', 'code']

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.label)



from time import time


try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle as pickle


class BaseCache(object):
    """
    Base class for the cache system. All the cache systems will
    implement this API or a superset of it.

    :param default_timeout: the default timeout that is used if no
                            timeout is specified on :meth:`set`.
    """
    def __init__(self, default_timeout=300, **kwargs):
        self.default_timeout = default_timeout

    def get(self, key):
        return None

    def set(self, key, value):
        pass

    def delete(self, key):
        pass

    def clear(self):
        pass

    def count(self):
        raise NotImplementedError

    def is_cached(self, key):
        return False


class NullCache(BaseCache):
    """
    A cache that doesn't cache.
    """


class SimpleCache(BaseCache):
    """
    Simply memory cache for single process environments. This class
    exists mainly for a development server and is not 100% thread safe.

    :param threshold: the maximum number of items the cache stores
                      before it starts evicting keys.
    """
    def __init__(self, threshold=500, default_timeout=300, **kwargs):
        super(SimpleCache, self).__init__(default_timeout, **kwargs)
        self._cache = {}
        self.threshold = threshold

    def _prune(self):
        if len(self._cache) > self.threshold:
            now = time()
            # Prune expired keys, or every few keys.
            for idx, (key, (expires, _)) in enumerate(self._cache.items()):
                if expires <= now or idx % 3 == 0:
                    self._cache.pop(key, None)

    def get(self, key):
        expires, value = self._cache.get(key, (0, None))
        if expires > time():
            return pickle.loads(value)

    def set(self, key, data_dict, timeout=None):
        if timeout is None:
            timeout = self.default_timeout
        self._prune()
        self._cache[key] = (time() + timeout, pickle.dumps(data_dict,
                            pickle.HIGHEST_PROTOCOL))

    def delete(self, key):
        return self._cache.pop(key, None)

    def clear(self):
        self._cache.clear()

    def count(self):
        return len(self._cache)

    def is_cached(self, key):
        return key in self._cache


class RedisCache(BaseCache):
    """Uses the Redis key-value store as a cache backend."""

    _connection_pool_cache = {}

    def __init__(self, host='localhost', port=6379, password=None,
                 db=0, default_timeout=300, key_prefix='', **kwargs):
        super(RedisCache, self).__init__(default_timeout, **kwargs)
        self.key_prefix = key_prefix
        self._client = self._get_client(host, port, password, db)

    @classmethod
    def _get_client(cls, host, port, password, db):
        """
        Retrieves a connection pool from a class-local cache (or creates
        it if necessary), returns a Redis client instance that uses that
        pool.
        """
        try:
            import redis
        except ImportError:
            raise RuntimeError('no redis module found')
        else:
            cls.REDIS_VERSION = redis.__version__

        # connection pool cache key
        credentials = (host, port, password, db)
        # retrieve or create a pool
        pool = cls._connection_pool_cache.get(
            credentials,
            redis.ConnectionPool(
                host=host,
                port=port,
                password=password,
                db=db,
            ),
        )
        cls._connection_pool_cache[credentials] = pool
        return redis.Redis(connection_pool=pool)

    def load_object(self, value):
        """The reversal of `dump_object`. This might be called with `None`."""
        if value is None:
            return None
        return pickle.loads(value)

    def dump_object(self, value):
        return pickle.dumps(value)

    def get(self, key):
        return self.load_object(self._client.get(self.key_prefix + key))

    def set(self, key, data_dict, timeout=None):
        if timeout is None:
            timeout = self.default_timeout
        data = self.dump_object(data_dict)
        # redis 3.x.x requires the order of arguments in setex to be:
        # (name, time, value)
        if self.REDIS_VERSION >= '3.0.0':
            return self._client.setex(self.key_prefix + key, timeout, data)
        # redis < 3.0.0
        return self._client.setex(self.key_prefix + key, data, timeout)

    def delete(self, key):
        return self._client.delete(self.key_prefix + key)

    def clear(self):
        cache_keys = self._client.keys('{}*'.format(self.key_prefix))
        # Python 2 and 3
        # inefficient on Python 2 to list a map
        list(map(self._client.delete, cache_keys))

    def info(self):
        return self._client.info()

    def is_cached(self, key):
        return self._client.exists(self.key_prefix + key)



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.resources.base import Resource
from gapipy.models.base import BaseModel

from .details import DossierDetail
from .dossier_features import DossierFeature


class MealBudget(BaseModel):
    _as_is_fields = [
        'id',
        'quality',
        'meal_type',
        'amount',
    ]


class CountryDossier(Resource):
    _resource_name = 'country_dossiers'

    _as_is_fields = [
        'id',
        'href',
        'type',
        'flags',
        'name',
        'publish_state',
    ]

    _date_time_fields_local = [
        'date_created',
        'date_last_modified',
    ]

    _resource_fields = [
        ('country', 'Country'),
        ('dossier_segment', 'DossierSegment'),
    ]

    _model_collection_fields = [
        ('details', DossierDetail),
        ('features', DossierFeature),
        ('meal_budgets', MealBudget),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.resources.base import Resource
from gapipy.models.base import BaseModel


class Requirement(BaseModel):
    _as_is_fields = ['code']


class RequirementSet(BaseModel):
    _as_is_fields = ['name', 'code']
    _model_collection_fields = [
        ('COMPLETE', Requirement),
        ('INCOMPLETE', Requirement),
    ]


class Checkin(Resource):
    _resource_name = 'checkins'
    _is_listable = True

    _as_is_fields = [
        'id',
        'href',
        'status',
    ]

    _date_fields = ['expires']
    _resource_fields = [
        ('booking', 'Booking'),
        ('customer', 'Customer'),
    ]

    _model_collection_fields = [
        ('requirements', RequirementSet),
    ]



# -*- coding: utf-8 -*-
# Python 2 and 3
from __future__ import unicode_literals

import datetime
import sys
from unittest import TestCase

from mock import patch

from gapipy.client import Client
from gapipy.query import Query
from gapipy.models import DATE_FORMAT, AccommodationRoom
from gapipy.resources import (
    ActivityDossier,
    Departure,
    Itinerary,
    Promotion,
    Tour,
    TourDossier,
)
from gapipy.resources.base import Resource
from gapipy.resources.tour.itinerary import ValidDuringRange

from .fixtures import DUMMY_DEPARTURE, PPP_TOUR_DATA, PPP_DOSSIER_DATA


class ResourceTestCase(TestCase):

    def setUp(self):
        self.client = Client()

    def test_to_dict(self):
        t = Tour(PPP_TOUR_DATA, client=self.client)
        self.assertEqual(t.to_dict(), PPP_TOUR_DATA)

    def test_to_dict_datetimes(self):
        class DatetimeResource(Resource):
            _as_is_fields = ['id']
            _date_fields = ['date_field']
            _date_time_fields_utc = ['date_field_utc']

        resource = DatetimeResource({
            'id': 1,
            'date_field': '2013-02-18',
            'date_field_utc': '2013-02-18T18:17:20Z',
        }, client=self.client)
        data = resource.to_dict()
        self.assertEqual(data['date_field'], '2013-02-18')
        self.assertEqual(data['date_field_utc'], '2013-02-18T18:17:20Z')

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_DOSSIER_DATA)
    def test_instantiate_from_raw_data(self, mock_request):
        t = Tour(PPP_TOUR_DATA, client=self.client)
        self.assertIsInstance(t, Tour)
        self.assertEqual(t.product_line, 'PPP')
        self.assertIsInstance(t.departures_start_date, datetime.date)
        self.assertIsInstance(t.tour_dossier, TourDossier)
        self.assertIsInstance(t.departures, Query)

    @patch('gapipy.request.APIRequestor._request')
    def test_populate_stub(self, mock_fetch):
        mock_fetch.return_value = {
            'id': 1,
            'product_line': 'PPP',
            'href': '/1',
            'departures_start_date': '2013-01-01',
            'departures_end_date': '2014-01-01',
            'tour_dossier': {
                'id': 1,
            },
            'departures': [
                {'id': 1},
                {'id': 2},
            ],
        }
        data = {'id': 1, 'product_line': 'PPP'}

        t = Tour(data, client=self.client, stub=True)
        self.assertTrue(t.is_stub)
        self.assertEqual(t.id, 1)

        self.assertTrue(isinstance(t.tour_dossier, TourDossier))
        self.assertTrue(isinstance(t.departures, Query))

        # Force a fetch.
        assert t.departures_start_date

        mock_fetch.assert_called_once()
        self.assertFalse(t.is_stub)
        self.assertEqual(
            t.departures_start_date,
            datetime.datetime.strptime('2013-01-01', DATE_FORMAT).date()
        )
        self.assertEqual(
            t.departures_end_date,
            datetime.datetime.strptime('2014-01-01', DATE_FORMAT).date()
        )

    def test_model_fields(self):
        from gapipy.models.base import BaseModel

        class Bar(BaseModel):
            _as_is_fields = ['id']
            _date_fields = ['date']

        class Foo(Resource):
            _model_fields = [('bar', Bar)]

        data = {
            'bar': {
                'id': 1,
                'date': '2013-01-01',
            }
        }
        f = Foo(data, client=self.client)

        self.assertEqual(f.to_dict(), {
            'bar': {'id': 1, 'date': '2013-01-01'}
        })
        self.assertIsInstance(f.bar, Bar)
        self.assertEqual(f.bar.id, 1)
        self.assertEqual(f.bar.date, datetime.date(2013, 0o1, 0o1))

    def test_null_model_fields(self):
        from gapipy.models.base import BaseModel

        class Bar(BaseModel):
            _as_is_fields = ['id']
            _date_fields = ['date']

        class Foo(Resource):
            _model_fields = [('bar', Bar)]

        data = {
            'bar': None
        }
        f = Foo(data, client=self.client)

        self.assertEqual(f.bar, None)


class AccommodationRoomTestCase(TestCase):

    def setUp(self):
        self.client = Client()

    def test_room_class_is_set_properly(self):
        data = {
            'code': 1234,
            'min_nights': 1,
            'max_nights': 2,
            'price_bands': [],
            'class': 'Standard',
        }
        room = AccommodationRoom(data, client=self.client)
        self.assertEqual(room.room_class, 'Standard')

    def test_room_class_is_set_properly_when_cached(self):
        data = {
            'code': 1234,
            'min_nights': 1,
            'max_nights': 2,
            'price_bands': [],
            'room_class': 'Standard',
        }
        room = AccommodationRoom(data, client=self.client)
        self.assertEqual(room.room_class, 'Standard')


class PromotionTestCase(TestCase):
    @patch('gapipy.request.APIRequestor._request')
    def test_product_type_is_set_properly(self, mock_request):
        data = {
            'products': [
                {
                    'type': 'departures',
                    'sub_type': 'Tour',
                    'id': 1,
                }
            ],
        }

        promotion = Promotion(data, client=Client())
        product = promotion.products[0]
        self.assertEqual(product.type, 'departures')
        self.assertEqual(product.to_dict(), {
            'id': 1,
            'type': 'departures',
            'sub_type': 'Tour',
        })

        # Ensure when the product implicitly fetched it makes a request, and
        # retains the type fields.
        self.assertEqual(mock_request.call_count, 0)
        mock_request.return_value = {
            'finish_address': {
                'country': {
                    'name': 'Australia',
                }
            }
        }
        self.assertEqual(product.finish_address.country.name, 'Australia')
        self.assertEqual(product.type, 'departures')


class PricePromotionTestCase(TestCase):
    def test_fake_amount_is_set_properly(self):
        departure = Departure(DUMMY_DEPARTURE, client=Client())
        prices = departure.rooms[0].price_bands[0].prices
        for price in prices:
            promotion = price.promotions[0].to_dict()
            self.assertTrue('amount' in promotion)


class DepartureAddonTestCase(TestCase):
    """
    Test that the departures.addons.product instances get appropriate resource
    types.

    This is a regression test for some erroneous behaviour of the AddOn model.

    The AddOn._resource_fields list was defined at the class level, and as-such
    it was shared among intances of that class. The effect is that if you have
    three addons, they will all be using the same model for their "product"
    even when those products are different type (e.g. "accommodations" versus
    "activities" versus "transports" versus "single_supplements" etc.)

    Desired behaviour is that each AddOn.product is represented by a model
    appropriate for its type.
    """
    def test_departure_addon_product_types(self):
        departure = Departure(DUMMY_DEPARTURE, client=Client())

        for addon in departure.addons:

            # Product should use a resource class that matches with its type
            self.assertEqual(addon.product._resource_name, addon.product.type)

            # Each addon instance should have only one resource class listed
            # for it's "product" field
            product_resource_fields = [
                (field, resource_class)
                for (field, resource_class) in addon._resource_fields
                if field == 'product']
            self.assertEqual(len(product_resource_fields), 1)


def build_vrange(start, end):
    """
    Returns a ValidDuringRange instance, given a start and end `date`

    We take care of formatting the `date` object as a string, and we just
    set the `client` argument to `None` because we don't need access to the
    API to test the validity functionality.
    """
    return ValidDuringRange({
        'start_date': start.isoformat() if start else None,
        'end_date': end.isoformat() if end else None,
    }, None)


class ValidDuringRangeTestCase(TestCase):
    """ Tests the date-range-checking logic on
    resources.tour.itinerary.ValidDuringRange
    """

    def test_is_valid_during_range__start(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # We'll be testing with a range that defines a start-date and no end...
        test_range = (today, None)

        # Check a range that falls entirely before our test_range
        self.assertFalse(
            build_vrange(today - (2 * a_week), today - a_week).is_valid_during_range(*test_range))

        # Check a range that falls before and overlapping our test_range
        self.assertTrue(
            build_vrange(today - a_week, today + a_week).is_valid_during_range(*test_range))

        # Check a range contained by our test_range
        self.assertTrue(
            build_vrange(today + a_week, today + (2 * a_week)).is_valid_during_range(*test_range))

        # Some cases we cannot check, because our test_range has no end:
        # - a range that contains our test_range
        # - a range that overlaps and falls after our test_range
        # - a range that falls entirely after our test_range

    def test_is_valid_during_range__end(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # We'll be testing with a range that defines an end-date and no start...
        test_range = (None, today)

        # Check a range contained by our test_range
        self.assertTrue(
            build_vrange(today - (2 * a_week), today - a_week).is_valid_during_range(*test_range))

        # Check a range that overlaps and falls after our test_range
        self.assertTrue(
            build_vrange(today - a_week, today + a_week).is_valid_during_range(*test_range))

        # Check a range that falls entirely after our test_range
        self.assertFalse(
            build_vrange(today + a_week, today + (2 * a_week)).is_valid_during_range(*test_range))

        # Some cases we cannot check, because our test_range has no start:
        # - a range that falls entirely before our test_range
        # - a range that falls before and overlapping our test_range
        # - a range that contains our test_range

    def test_is_valid_during_range__start_end(self):
        today = datetime.date.today()
        a_day = datetime.timedelta(1)
        a_week = datetime.timedelta(7)

        # We'll be testing with a range that defines both a start and end date
        test_range = (today, today + a_week)

        # Check ranges that fall entirely before our test_range
        self.assertFalse(
            build_vrange(today - (2 * a_week), today - a_week).is_valid_during_range(*test_range))
        self.assertFalse(
            build_vrange(None, today - a_week).is_valid_during_range(*test_range))

        # Check a range that falls before and overlapping our test_range
        self.assertTrue(
            build_vrange(today - (2 * a_week), today).is_valid_during_range(*test_range))

        # Check a range that contains our test_range
        self.assertTrue(
            build_vrange(today - (2 * a_week), today + (2 * a_week)).is_valid_during_range(*test_range))

        # Check a range contained by our test_range
        self.assertTrue(
            build_vrange(today + a_day, today + a_week - a_day).is_valid_during_range(*test_range))

        # Check a range that overlaps and falls after our test_range
        self.assertTrue(
            build_vrange(today + a_day, today + (2 * a_week)).is_valid_during_range(*test_range))

        # Check ranges that fall entirely after our test_range
        self.assertFalse(
            build_vrange(today + (2 * a_week), today + (3 * a_week)).is_valid_during_range(*test_range))
        self.assertFalse(
            build_vrange(today + (2 * a_week), None).is_valid_during_range(*test_range))

    def test_is_valid_during_range__none(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # We'll be testing with a range that defines neither start or end
        # date... this should be equivalent to the is_valid_sometime call..
        test_range = (None, None)

        # Check a range that contains our test_range
        self.assertTrue(
            build_vrange(None, None).is_valid_during_range(*test_range))

        # Check a range contained by our test_range
        self.assertTrue(
            build_vrange(today, None).is_valid_during_range(*test_range))
        self.assertTrue(
            build_vrange(None, today).is_valid_during_range(*test_range))
        self.assertTrue(
            build_vrange(today - a_week, today).is_valid_during_range(*test_range))

        # Some cases we cannot check, because our test_range is (None, None):
        # - a range that falls entirely before our test_range
        # - a range that falls before and overlapping our test_range
        # - a range that overlaps and falls after our test_range
        # - a range that falls entirely after our test_range

    def test_is_valid_during_range__invalid(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # We'll be testing with a range that starts after it ends (so, NOTHING
        # should be valid in this range)
        test_range = (today, today - a_week)

        self.assertFalse(
            build_vrange(today, None).is_valid_during_range(*test_range))
        self.assertFalse(
            build_vrange(None, today).is_valid_during_range(*test_range))
        self.assertFalse(
            build_vrange(today - a_week, today).is_valid_during_range(*test_range))

    def test_is_valid_on_or_after_date(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check a range that is valid ON the date in question (but not AFTER the date)
        self.assertTrue(
            build_vrange(today, today).is_valid_on_or_after_date(today))

        # Check a range that is valid AFTER the date in question (but not ON the date)
        self.assertTrue(
            build_vrange(today + a_week, None).is_valid_on_or_after_date(today))

        # Check some ranges that are valid before, on, AND after the date in question
        self.assertTrue(
            build_vrange(None, None).is_valid_on_or_after_date(today))
        self.assertTrue(
            build_vrange(today - a_week, None).is_valid_on_or_after_date(today))

        # Check some ranges (closed and half-open) that are not valid after or on the date
        self.assertFalse(
            build_vrange(today - (2 * a_week), today - a_week).is_valid_on_or_after_date(today))
        self.assertFalse(
            build_vrange(None, today - a_week).is_valid_on_or_after_date(today))

    def test_is_valid_on_or_before_date(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check a range that is valid ON the date in question (but not BEFORE the date)
        self.assertTrue(
            build_vrange(today, today).is_valid_on_or_before_date(today))

        # Check a range that is valid BEFORE the date in question (but not ON the date)
        self.assertTrue(
            build_vrange(None, today - a_week).is_valid_on_or_before_date(today))

        # Check some ranges that are valid before, on, AND after the date in question
        self.assertTrue(
            build_vrange(None, None).is_valid_on_or_before_date(today))
        self.assertTrue(
            build_vrange(today - a_week, None).is_valid_on_or_before_date(today))

        # Check some ranges (closed and half-open) that are not valid before or on the date
        self.assertFalse(
            build_vrange(today + a_week, today + (2 * a_week)).is_valid_on_or_before_date(today))
        self.assertFalse(
            build_vrange(today + a_week, None).is_valid_on_or_before_date(today))

    def test_is_valid_on_date(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        invalid_range_closed = build_vrange(today - (2 * a_week), today - a_week)
        self.assertFalse(invalid_range_closed.is_valid_on_date(today))

        invalid_range_half_open = build_vrange(today + (2 * a_week), None)
        self.assertFalse(invalid_range_half_open.is_valid_on_date(today))

        valid_range_closed = build_vrange(today - (2 * a_week), today + a_week)
        self.assertTrue(valid_range_closed.is_valid_on_date(today))

        valid_range_half_open = build_vrange(None, today + a_week)
        self.assertTrue(valid_range_half_open.is_valid_on_date(today))

        valid_range_open = build_vrange(None, None)
        self.assertTrue(valid_range_open.is_valid_on_date(today))

    def test_is_valid_sometime(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # A range that starts after it finishes is never valid
        invalid_range = build_vrange(today + a_week, today - a_week)
        self.assertFalse(invalid_range.is_valid_sometime())

        # Otherwise, it's valid sometime if there is a start...
        self.assertTrue(build_vrange(today, None).is_valid_sometime())

        # ... or end ...
        self.assertTrue(build_vrange(None, today).is_valid_sometime())

        # ... or both (where start >= end)
        self.assertTrue(build_vrange(today + a_week, today + a_week).is_valid_sometime())


class ItineraryValidityTestCase(TestCase):
    """ Tests the date-range-checking logic on
    resources.tour.itinerary.Itinerary
    """
    def build_itin(self, ranges):
        """
        Returns an Itinerary instance, given a list of (start, end) `date`
        tuples to be used to populate the list of `valid_during_ranges`.
        """
        ranges = ranges or []
        return Itinerary({
            'id': 'who cares',
            'valid_during_ranges': [
                {
                    'start_date': start.isoformat() if start else None,
                    'end_date': end.isoformat() if end else None,
                } for (start, end) in ranges
            ],
        }, None)

    def test_is_expired(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check no ranges
        self.assertTrue(self.build_itin(None).is_expired())

        # Check one expired
        self.assertTrue(
            self.build_itin([
                (today - (2 * a_week), today - a_week),
            ]).is_expired()
        )

        # Check one expired, one valid
        self.assertFalse(
            self.build_itin([
                (today - (2 * a_week), today - a_week),
                (today, today + a_week),
            ]).is_expired()
        )

        # Check one valid
        self.assertFalse(
            self.build_itin([
                (today, today + a_week),
            ]).is_expired()
        )

    def test_is_valid_during_range(self):
        today = datetime.date.today()
        a_day = datetime.timedelta(1)
        a_week = datetime.timedelta(7)

        test_range = (today + a_day, today + a_day + a_week)

        # Check no ranges
        self.assertFalse(self.build_itin(None).is_valid_during_range(*test_range))

        # Check one invalid
        self.assertFalse(
            self.build_itin([
                (today - a_week, today),
            ]).is_valid_during_range(*test_range)
        )

        # Check one invalid, one valid
        self.assertTrue(
            self.build_itin([
                (today - a_week, today),
                (today, today + a_week),
            ]).is_valid_during_range(*test_range)
        )

        # Check one valid
        self.assertTrue(
            self.build_itin([
                (today, today + a_week),
            ]).is_valid_during_range(*test_range)
        )

    def test_is_valid_on_or_after_date(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check no ranges
        self.assertFalse(self.build_itin(None).is_valid_on_or_after_date(today))

        # Check one invalid
        self.assertFalse(
            self.build_itin([
                (today - (2 * a_week), today - a_week),
            ]).is_valid_on_or_after_date(today)
        )

        # Check one invalid, one valid
        self.assertTrue(
            self.build_itin([
                (today + a_week, today + (2 * a_week)),
                (today - (2 * a_week), today - a_week),
            ]).is_valid_on_or_after_date(today)
        )

        # Check one valid
        self.assertTrue(
            self.build_itin([
                (today + a_week, today + (2 * a_week)),
            ]).is_valid_on_or_after_date(today)
        )

    def test_is_valid_on_or_before_date(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check no ranges
        self.assertFalse(self.build_itin(None).is_valid_on_or_before_date(today))

        # Check one invalid
        self.assertFalse(
            self.build_itin([
                (today + a_week, today + (2 * a_week)),
            ]).is_valid_on_or_before_date(today)
        )

        # Check one invalid, one valid
        self.assertTrue(
            self.build_itin([
                (today + a_week, today + (2 * a_week)),
                (today - (2 * a_week), today - a_week),
            ]).is_valid_on_or_before_date(today)
        )

        # Check one valid
        self.assertTrue(
            self.build_itin([
                (today - (2 * a_week), today - a_week),
            ]).is_valid_on_or_before_date(today)
        )

    def test_is_valid_on_date(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check no ranges
        self.assertFalse(self.build_itin(None).is_valid_on_date(today))

        # Check one invalid
        self.assertFalse(
            self.build_itin([
                (today - (2 * a_week), today - a_week),
            ]).is_valid_on_date(today)
        )

        # Check one invalid, one valid
        self.assertTrue(
            self.build_itin([
                (today - (2 * a_week), today - a_week),
                (today, today + a_week),
            ]).is_valid_on_date(today)
        )

        # Check one valid
        self.assertTrue(
            self.build_itin([
                (today, today + a_week),
            ]).is_valid_on_date(today)
        )

    def test_is_valid_sometime(self):
        today = datetime.date.today()
        a_week = datetime.timedelta(7)

        # Check no ranges
        self.assertFalse(self.build_itin(None).is_valid_sometime())

        # Check one invalid (note start is after finish)
        self.assertFalse(
            self.build_itin([
                (today - a_week, today - (2 * a_week)),
            ]).is_valid_sometime()
        )

        # Check one invalid, one valid
        self.assertTrue(
            self.build_itin([
                (today - a_week, today - (2 * a_week)),
                (today, today + a_week),
            ]).is_valid_sometime()
        )

        # Check one valid
        self.assertTrue(
            self.build_itin([
                (today, today + a_week),
            ]).is_valid_sometime()
        )



# -*- coding: utf-8 -*-
# Python 2 and 3
from __future__ import unicode_literals

from ..base import Resource
from ...models.base import BaseModel


MAP_TYPE = 'MAP'
OTHER_TYPE = 'OTHER'

IMAGE_TYPES = (
    MAP_TYPE,
    OTHER_TYPE,
)

class ImageFile(BaseModel):
    _as_is_fields = [
        'url', 'data', 'mime_type', 'exif',
    ]


class Image(Resource):

    _resource_name = 'images'

    _as_is_fields = [
        'id', 'href', 'modification', 'description', 'keywords',
        'attribution', 'channels', 'variations', 'type',
    ]
    _date_time_fields_local = ['date_created', 'date_last_modified']
    _resource_fields = [('original', 'Image')]
    _model_fields = [('file', ImageFile)]



# -*- coding: utf-8 -*-
# Python 2 and 3
from __future__ import unicode_literals

from ..base import Resource



class Video(Resource):

    _resource_name = 'videos'

    _as_is_fields = [
        'id',
        'href',
        'url',
        'code',
        'source',
        'title',
        'description',
        'image',
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models import PriceBand
from gapipy.resources.booking_company import BookingCompany
from gapipy.resources.product import Product


class Merchandise(Product):

    _resource_name = 'merchandise'

    _is_listable = False

    _as_is_fields = [
        'id',
        'availability',
        'name',
        'product_line',
        'sku',
        'type',
        'sub_type',
        'href',
    ]
    _date_fields = [
        'start_date',
        'finish_date',
    ]
    _date_time_fields_utc = [
        'date_created',
        'date_last_modified',
    ]
    _model_collection_fields = [
        ('price_bands', PriceBand),
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.resources.base import Resource
from gapipy.utils import enforce_string_type

from .country import Country


class Nationality(Resource):

    _resource_name = 'nationalities'

    _as_is_fields = [
        'id',
        'href',
        'name',
    ]

    _resource_fields = [
        ('country', Country),
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)



# Python 2 and 3
from __future__ import unicode_literals

from .base import Resource


class BookingCompany(Resource):

    _resource_name = 'booking_companies'

    _as_is_fields = [
        'id',
        'href',
        'name',
    ]



# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models.base import BaseModel
from gapipy.resources.base import Resource
from gapipy.utils import get_resource_class_from_resource_name

from .service import Service


class LastModified(BaseModel):
    _as_is_fields = [
        'name',
        'user_type',
    ]
    _date_time_fields_utc = ['date']


class ResourceField(BaseModel):
    _as_is_fields = [
        'field',
        'required',
        'value',
        'viewable_by',
    ]

    @property
    def _resource_fields(self):
        # extract the resource_name from the raw data
        # and use it to return the appropriate resource
        # class
        resource_name = self._raw_data.get('resource', {}).get('resource_name')
        if resource_name:
            try:
                resource_class = get_resource_class_from_resource_name(resource_name)
                return [
                    ('resource', resource_class),
                ]
            except:
                pass
        # fall back
        return []


class Requirement(Resource):
    _resource_name = 'requirements'

    _as_is_fields = [
        'id',
        'href',
        'type',
        'applied_overrides',
        'code',
        'editable_by',
        'flags',
        'name',
        'override_reasons',
        'status',
    ]

    _date_fields = [
        'complete_by_date',
    ]

    _model_fields = [
        ('last_modified', LastModified),
    ]

    _resource_fields = [
        ('booking', 'Booking'),
        ('customer', 'Customer'),
        ('requirement_set', 'RequirementSet'),
    ]

    _model_collection_fields = [
        ('services', Service),
        ('resource_fields', ResourceField),
    ]


class RequirementSet(Resource):
    _resource_name = 'requirement_sets'

    _is_listable = True

    _as_is_fields = [
        'id',
        'href',
        'code',
        'name',
    ]

    _date_fields = [
        'complete_by_date',
    ]

    _model_collection_fields = [
        ('requirements', Requirement),
    ]

    _resource_fields = [
        ('booking', 'Booking'),
    ]




# Python 2 and 3
from __future__ import unicode_literals

from gapipy.models import Address, AccommodationRoom
from gapipy.utils import enforce_string_type
from gapipy.resources.booking_company import BookingCompany
from gapipy.resources.dossier import AccommodationDossier
from gapipy.resources.product import Product


class Accommodation(Product):

    _resource_name = 'accommodations'

    _as_is_fields = [
        'id',
        'href',
        'name',
        'product_line',
        'sku',
        'type',
        'sub_type',
        'phone_numbers',
    ]
    _date_time_fields_utc = [
        'date_created',
        'date_last_modified',
    ]
    _model_fields = [
        ('address', Address),
    ]
    _model_collection_fields = [
        ('booking_companies', BookingCompany),
        ('rooms', AccommodationRoom),
    ]
    _resource_fields = [
        ('accommodation_dossier', AccommodationDossier),
    ]

    @enforce_string_type
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)



from .base import BaseModel

class DossierFeatureParent(BaseModel):
    _as_is_fields = ['id', 'label', 'code', 'description']

class DossierFeatureChild(BaseModel):
    _as_is_fields = ['id', 'label', 'code', 'description']



"""These tests mostly ensure that any change in the shape of the data
returned by the API is detected.
"""
import sys
from unittest import TestCase

from nose.plugins.attrib import attr
from nose_parameterized import parameterized

from gapipy.client import Client
from gapipy.resources import (
    Accommodation,
    AccommodationDossier,
    Activity,
    ActivityDossier,
    Airport,
    Checkin,
    Continent,
    Country,
    CountryDossier,
    Departure,
    DepartureComponent,
    DossierFeature,
    DossierSegment,
    Feature,
    FeatureCategory,
    Image,
    Itinerary,
    Language,
    Merchandise,
    OverrideReason,
    Place,
    PlaceDossier,
    Promotion,
    ServiceLevel,
    SingleSupplement,
    State,
    Timezone,
    Tour,
    TourCategory,
    TourDossier,
    Transport,
    TransportDossier,
    available_public_resources,
)


@attr('integration')
class TourTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tour = Client().tours.get(21715)

    def setUp(self):
        self.tour = TourTestCase.tour


@attr('integration')
class TourDossierTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dossier = Client().tour_dossiers.get(24515)

    def setUp(self):
        self.dossier = TourDossierTestCase.dossier

    def test_get_image_url(self):
        url = self.dossier.get_map_url()

        if sys.version_info.major < 3:
            # Python 2
            self.assertIsInstance(url, basestring)
        else:
            # Python 3
            self.assertIsInstance(url, str)

    def test_get_countries(self):
        countries = self.dossier.get_visited_countries()
        self.assertIsInstance(countries, list)

    def test_get_trip_detail(self):
        detail = self.dossier.get_trip_detail('Highlights')
        if sys.version_info.major < 3:
            # Python 2
            self.assertIsInstance(detail, basestring)
        else:
            # Python 3
            self.assertIsInstance(detail, str)

    def test_get_category_name(self):
        category = self.dossier.get_category_name('Trip Type')
        if sys.version_info.major < 3:
            # Python 2
            self.assertIsInstance(category, basestring)
        else:
            # Python 3
            self.assertIsInstance(category, str)


@attr('integration')
class LiveAPITestCase(TestCase):

    resources = [
        (Accommodation, 54),
        (AccommodationDossier, 5856),
        (Activity, 4486),
        (ActivityDossier, 9695),
        (Airport, 25553),
        (Checkin, 'q8YJAtaV9x'),
        (Continent, 'NA'),
        (Country, 'CA'),
        (CountryDossier, 10382),
        (Departure, 419504),
        (DepartureComponent, '55_3983802'),
        (DossierFeature, 4),
        (DossierSegment, 19),
        (Feature, 214),
        (FeatureCategory, 4),
        (Image, 4321),
        (Itinerary, 978),
        (Language, '250'),
        (Merchandise, 90867),
        (OverrideReason, 4),
        (Place, 2039527),
        (PlaceDossier, 711),
        (Promotion, 18336),
        (ServiceLevel, 3),
        (SingleSupplement, 'T419504'),
        (State, 'CA-ON'),
        (Timezone, 4),
        (Tour, 21715),
        (TourCategory, 15),
        (TourDossier, 24515),
        (Transport, 298),
        (TransportDossier, 51),
    ]

    def test_all_public_resources_are_tested(self):
        """This a bit meta, this is a test to ensure that all available
        (public) resources are included in this integration test case.
        """
        tested_resources = [r.__name__ for (r, pk) in self.resources]
        self.assertEqual(sorted(available_public_resources), sorted(tested_resources))

    @parameterized.expand(resources)
    def test_fetch_resource(self, resource, resource_id):

        # Test if a resource can be parsed properly. This is useful to catch
        # changes in the API
        api = Client()
        query = getattr(api, resource._resource_name)
        instance = query.get(resource_id)

        # Test that we take into account all of the fields of the json data
        # returned by the API (this only checks for the top-level fields, not
        # those of nested objects)
        self.assertEqual(
            sorted(instance._allowed_fields()),
            sorted(instance._raw_data.keys())
        )
